# The MIT License (MIT)
#
# Copyright (c) 2014-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import termcolor
import os.path
import barectf
import barectf.config_parse_common as barectf_config_parse_common
import barectf.argpar as barectf_argpar
from typing import Any, List, Iterable, NoReturn
import typing
from barectf.typing import Index, Count
import sys
import os


# Colors and prints the error message `msg` and exits with status code
# 1.
def _print_error(msg: str) -> NoReturn:
    termcolor.cprint('Error: ', 'red', end='', file=sys.stderr)
    termcolor.cprint(msg, 'red', attrs=['bold'], file=sys.stderr)
    sys.exit(1)


# Pretty-prints the barectf configuration error `exc` and exits with
# status code 1.
def _print_config_error(exc: barectf._ConfigurationParseError) -> NoReturn:
    # reverse: most precise message comes last
    for ctx in reversed(exc.context):
        msg = ''

        if ctx.message is not None:
            msg = f' {ctx.message}'

        color = 'red'
        termcolor.cprint(f'{ctx.name}', color, attrs=['bold'], file=sys.stderr, end='')
        termcolor.cprint(':', color, file=sys.stderr, end='')
        termcolor.cprint(msg, color, file=sys.stderr)

    sys.exit(1)


# Pretty-prints the unknown exception `exc`.
def _print_unknown_exc(exc: Exception) -> NoReturn:
    import traceback

    traceback.print_exc()
    _print_error(f'Unknown exception: {exc}')


# Finds and returns all the option items in `items` having the long name
# `long_name`.
def _find_opt_items(items: Iterable[barectf_argpar._Item],
                    long_name: str) -> List[barectf_argpar._OptItem]:
    ret_items: List[barectf_argpar._OptItem] = []

    for item in items:
        if type(item) is barectf_argpar._OptItem:
            item = typing.cast(barectf_argpar._OptItem, item)

            if item.descr.long_name == long_name:
                ret_items.append(item)

    return ret_items


# Returns:
#
# For an option item without an argument:
#     `True`.
#
# For an option item with an argument:
#     Its argument.
#
# Uses the last option item having the long name `long_name` found in
# `items`.
#
# Returns `default` if there's no such option item.
def _opt_item_val(items: Iterable[barectf_argpar._Item], long_name: str,
                  default: Any = None) -> Any:
    opt_items = _find_opt_items(items, long_name)

    if len(opt_items) == 0:
        return default

    opt_item = opt_items[-1]

    if opt_item.descr.has_arg:
        return opt_item.arg_text

    return True


class _CliError(Exception):
    pass


def _cfg_file_path_from_parse_res(parse_res: barectf_argpar._ParseRes) -> str:
    cfg_file_path = None

    for item in parse_res.items:
        if type(item) is barectf_argpar._NonOptItem:
            if cfg_file_path is not None:
                raise _CliError('Multiple configuration file paths provided')

            cfg_file_path = typing.cast(barectf_argpar._NonOptItem, item).text

    if cfg_file_path is None:
        raise _CliError('Missing configuration file path')

    if not os.path.isfile(cfg_file_path):
        raise _CliError(f'`{cfg_file_path}` is not an existing, regular file')

    return cfg_file_path


# Returns a `_CfgCmdCfg` object from the command-line parsing results
# `parse_res`.
def _cfg_cmd_cfg_from_parse_res(parse_res: barectf_argpar._ParseRes) -> '_CfgCmdCfg':
    # check configuration file path
    cfg_file_path = _cfg_file_path_from_parse_res(parse_res)

    # inclusion directories (`--include-dir` option needs an argument)
    inclusion_dirs = typing.cast(List[str],
                                 [item.arg_text for item in _find_opt_items(parse_res.items, 'include-dir')])

    for dir in inclusion_dirs:
        if not os.path.isdir(dir):
            raise _CliError(f'`{dir}` is not an existing directory')

    inclusion_dirs.append(os.getcwd())

    # other options
    ignore_inclusion_file_not_found = _opt_item_val(parse_res.items, 'ignore-include-not-found',
                                                    False)

    return _CfgCmdCfg(cfg_file_path, inclusion_dirs, ignore_inclusion_file_not_found)


def _print_gen_cmd_usage():
    print('''Usage: barectf generate [--code-dir=DIR] [--headers-dir=DIR]
                        [--metadata-dir=DIR] [--prefix=PREFIX]
                        [--include-dir=DIR]... [--ignore-include-not-found]
                        CONFIG-FILE-PATH
       barectf generate --help

Generate the C source and CTF metadata stream files of a tracer from the
configuration file CONFIG-FILE-PATH.

Options:
  -c DIR, --code-dir=DIR        Write C source files to DIR instead of the CWD
  -H DIR, --headers-dir=DIR     Write C header files to DIR instead of the CWD
  -h, --help                    Show this help and quit
  --ignore-include-not-found    Continue to process the configuration file when
                                included files are not found
  -I DIR, --include-dir=DIR     Add DIR to the list of directories to be
                                searched for inclusion files
  -m DIR, --metadata-dir=DIR    Write the metadata stream file to DIR instead of
                                the CWD
  -p PREFIX, --prefix=PREFIX    Set the configuration prefix to PREFIX''')


# Returns a source and metadata stream file generating command object
# from the specific command-line arguments `orig_args`.
def _gen_cmd_cfg_from_args(orig_args: barectf_argpar.OrigArgs) -> '_GenCmd':
    # parse original arguments
    opt_descrs = [
        barectf_argpar.OptDescr('h', 'help'),
        barectf_argpar.OptDescr('c', 'code-dir', True),
        barectf_argpar.OptDescr('H', 'headers-dir', True),
        barectf_argpar.OptDescr('I', 'include-dir', True),
        barectf_argpar.OptDescr('m', 'metadata-dir', True),
        barectf_argpar.OptDescr('p', 'prefix', True),
        barectf_argpar.OptDescr(long_name='dump-config'),
        barectf_argpar.OptDescr(long_name='ignore-include-not-found'),
    ]
    res = barectf_argpar.parse(orig_args, opt_descrs)
    assert len(res.ingested_orig_args) == len(orig_args)

    # command help?
    if len(_find_opt_items(res.items, 'help')) > 0:
        _print_gen_cmd_usage()
        sys.exit()

    # get common configuration file command CLI configuration
    cfg_cmd_cfg = _cfg_cmd_cfg_from_parse_res(res)

    # directories
    c_source_dir = _opt_item_val(res.items, 'code-dir', os.getcwd())
    c_header_dir = _opt_item_val(res.items, 'headers-dir', os.getcwd())
    metadata_stream_dir = _opt_item_val(res.items, 'metadata-dir', os.getcwd())

    for dir in [c_source_dir, c_header_dir, metadata_stream_dir]:
        if not os.path.isdir(dir):
            raise _CliError(f'`{dir}` is not an existing directory')

    # other options
    dump_config = _opt_item_val(res.items, 'dump-config', False)
    v2_prefix = _opt_item_val(res.items, 'prefix')

    return _GenCmd(_GenCmdCfg(cfg_cmd_cfg.cfg_file_path, c_source_dir, c_header_dir,
                              metadata_stream_dir, cfg_cmd_cfg.inclusion_dirs,
                              cfg_cmd_cfg.ignore_inclusion_file_not_found, dump_config, v2_prefix))


def _show_effective_cfg_cmd_usage():
    print('''Usage: barectf show-effective-configuration [--include-dir=DIR]...
                                            [--ignore-include-not-found]
                                            [--indent-spaces=COUNT] CONFIG-FILE-PATH
       barectf show-effective-configuration --help

Print the effective configuration file for a the configuration file
CONFIG-FILE-PATH.

Options:
  -h, --help                    Show this help and quit
  --ignore-include-not-found    Continue to process the configuration file when
                                included files are not found
  -I DIR, --include-dir=DIR     Add DIR to the list of directories to be
                                searched for inclusion files
  --indent-spaces=COUNT         Use COUNT spaces at a time to indent YAML lines
                                instead of 2''')


# Returns an effective configuration showing command object from the
# specific command-line arguments `orig_args`.
def _show_effective_cfg_cfg_from_args(orig_args: barectf_argpar.OrigArgs) -> '_ShowEffectiveCfgCmd':
    # parse original arguments
    opt_descrs = [
        barectf_argpar.OptDescr('h', 'help'),
        barectf_argpar.OptDescr('I', 'include-dir', True),
        barectf_argpar.OptDescr(long_name='indent-spaces', has_arg=True),
        barectf_argpar.OptDescr(long_name='ignore-include-not-found'),
    ]
    res = barectf_argpar.parse(orig_args, opt_descrs)
    assert len(res.ingested_orig_args) == len(orig_args)

    # command help?
    if len(_find_opt_items(res.items, 'help')) > 0:
        _show_effective_cfg_cmd_usage()
        sys.exit()

    # get common configuration command CLI configuration
    cfg_cmd_cfg = _cfg_cmd_cfg_from_parse_res(res)

    # other options
    indent_space_count = _opt_item_val(res.items, 'indent-spaces', 2)

    try:
        indent_space_count = int(indent_space_count)
    except (ValueError, TypeError):
        raise _CliError(f'Invalid `--indent-spaces` option argument: `{indent_space_count}`')

    if indent_space_count < 1 or indent_space_count > 8:
        raise _CliError(f'Invalid `--indent-spaces` option argument (`{indent_space_count}`): expecting a value in [1, 8]')

    return _ShowEffectiveCfgCmd(_ShowEffectiveCfgCmdCfg(cfg_cmd_cfg.cfg_file_path,
                                                        cfg_cmd_cfg.inclusion_dirs,
                                                        cfg_cmd_cfg.ignore_inclusion_file_not_found,
                                                        Count(indent_space_count)))


def _show_cfg_version_cmd_usage():
    print('''Usage: barectf show-configuration-version CONFIG-FILE-PATH
       barectf show-configuration-version --help

Print the major version (2 or 3) of the configuration file CONFIG-FILE-PATH.

Options:
  -h, --help    Show this help and quit''')


# Returns a configuration version showing command object from the
# specific command-line arguments `orig_args`.
def _show_cfg_version_cfg_from_args(orig_args: barectf_argpar.OrigArgs) -> '_ShowCfgVersionCmd':
    # parse original arguments
    opt_descrs = [
        barectf_argpar.OptDescr('h', 'help'),
    ]
    res = barectf_argpar.parse(orig_args, opt_descrs)
    assert len(res.ingested_orig_args) == len(orig_args)

    # command help?
    if len(_find_opt_items(res.items, 'help')) > 0:
        _show_cfg_version_cmd_usage()
        sys.exit()

    # check configuration file path
    cfg_file_path = _cfg_file_path_from_parse_res(res)

    return _ShowCfgVersionCmd(_ShowCfgVersionCmdCfg(cfg_file_path))


def _print_general_usage():
    print('''Usage: barectf COMMAND COMMAND-ARGS
       barectf --help
       barectf --version

General options:
  -h, --help       Show this help and quit
  -V, --version    Show version and quit

Available commands:
  gen:
  generate:
    Generate the C source and CTF metadata stream files of a tracer from a
    configuration file.

  show-effective-configuration:
  show-effective-config:
  show-effective-cfg:
    Print the effective configuration file for a given configuration file and
    inclusion directories.

  show-configuration-version:
  show-config-version:
  show-cfg-version
    Print the major version of a given configuration file.

Run `barectf COMMAND --help` to show the help of COMMAND.''')


# Returns a command object from the command-line arguments `orig_args`.
#
# All the `orig_args` elements are considered.
def _cmd_from_args(orig_args: barectf_argpar.OrigArgs) -> '_Cmd':
    # We use our `argpar` module here instead of Python's `argparse`
    # because we need to support the two following use cases:
    #
    #     $ barectf config.yaml
    #     $ barectf generate config.yaml
    #
    # In other words, the default command is `generate` (for backward
    # compatibility reasons). The argument parser must not consider
    # `config.yaml` as being a command name.
    general_opt_descrs = [
        barectf_argpar.OptDescr('V', 'version'),
        barectf_argpar.OptDescr('h', 'help'),
    ]
    res = barectf_argpar.parse(orig_args, general_opt_descrs, False)

    # find command name, collecting preceding (common) option items
    cmd_from_args_funcs = {
        'generate': _gen_cmd_cfg_from_args,
        'gen': _gen_cmd_cfg_from_args,
        'show-effective-configuration': _show_effective_cfg_cfg_from_args,
        'show-effective-config': _show_effective_cfg_cfg_from_args,
        'show-effective-cfg': _show_effective_cfg_cfg_from_args,
        'show-configuration-version': _show_cfg_version_cfg_from_args,
        'show-config-version': _show_cfg_version_cfg_from_args,
        'show-cfg-version': _show_cfg_version_cfg_from_args,
    }
    general_opt_items: List[barectf_argpar._OptItem] = []
    cmd_first_orig_arg_index = None
    cmd_from_args_func = None

    for item in res.items:
        if type(item) is barectf_argpar._NonOptItem:
            item = typing.cast(barectf_argpar._NonOptItem, item)
            cmd_from_args_func = cmd_from_args_funcs.get(item.text)

            if cmd_from_args_func is None:
                cmd_first_orig_arg_index = item.orig_arg_index
            else:
                cmd_first_orig_arg_index = Index(item.orig_arg_index + 1)

            break
        else:
            assert type(item) is barectf_argpar._OptItem
            general_opt_items.append(typing.cast(barectf_argpar._OptItem, item))

    # general help?
    if len(_find_opt_items(general_opt_items, 'help')) > 0:
        _print_general_usage()
        sys.exit()

    # version?
    if len(_find_opt_items(general_opt_items, 'version')) > 0:
        print(f'barectf {barectf.__version__}')
        sys.exit()

    # create command object
    cmd_orig_args = orig_args[cmd_first_orig_arg_index:]

    if cmd_from_args_func is None:
        # default `generate` command
        return _gen_cmd_cfg_from_args(cmd_orig_args)
    else:
        return cmd_from_args_func(cmd_orig_args)


class _CmdCfg:
    pass


class _CfgCmdCfg(_CmdCfg):
    def __init__(self, cfg_file_path: str, inclusion_dirs: List[str],
                 ignore_inclusion_file_not_found: bool):
        self._cfg_file_path = cfg_file_path
        self._inclusion_dirs = inclusion_dirs
        self._ignore_inclusion_file_not_found = ignore_inclusion_file_not_found

    @property
    def cfg_file_path(self) -> str:
        return self._cfg_file_path

    @property
    def inclusion_dirs(self) -> List[str]:
        return self._inclusion_dirs

    @property
    def ignore_inclusion_file_not_found(self) -> bool:
        return self._ignore_inclusion_file_not_found


class _Cmd:
    def __init__(self, cfg: _CmdCfg):
        self._cfg = cfg

    @property
    def cfg(self) -> _CmdCfg:
        return self._cfg

    def exec(self):
        raise NotImplementedError


class _GenCmdCfg(_CfgCmdCfg):
    def __init__(self, cfg_file_path: str, c_source_dir: str, c_header_dir: str,
                 metadata_stream_dir: str, inclusion_dirs: List[str],
                 ignore_inclusion_file_not_found: bool, dump_config: bool, v2_prefix: str):
        super().__init__(cfg_file_path, inclusion_dirs, ignore_inclusion_file_not_found)
        self._c_source_dir = c_source_dir
        self._c_header_dir = c_header_dir
        self._metadata_stream_dir = metadata_stream_dir
        self._dump_config = dump_config
        self._v2_prefix = v2_prefix

    @property
    def c_source_dir(self) -> str:
        return self._c_source_dir

    @property
    def c_header_dir(self) -> str:
        return self._c_header_dir

    @property
    def metadata_stream_dir(self) -> str:
        return self._metadata_stream_dir

    @property
    def dump_config(self) -> bool:
        return self._dump_config

    @property
    def v2_prefix(self) -> str:
        return self._v2_prefix


# Source and metadata stream file generating command.
class _GenCmd(_Cmd):
    def exec(self):
        # create configuration
        try:
            with open(self.cfg.cfg_file_path) as f:
                if self.cfg.dump_config:
                    # print effective configuration file
                    print(barectf.effective_configuration_file(f, True, self.cfg.inclusion_dirs,
                                                               self.cfg.ignore_inclusion_file_not_found))

                    # barectf.configuration_from_file() reads the file again
                    # below: rewind.
                    f.seek(0)

                config = barectf.configuration_from_file(f, True, self.cfg.inclusion_dirs,
                                                         self.cfg.ignore_inclusion_file_not_found)
        except barectf._ConfigurationParseError as exc:
            _print_config_error(exc)
        except Exception as exc:
            _print_unknown_exc(exc)

        if self.cfg.v2_prefix is not None:
            # Override prefixes.
            #
            # For historical reasons, the `--prefix` option applies the
            # barectf 2 configuration prefix rules. Therefore, get the
            # equivalent barectf 3 prefixes first.
            v3_prefixes = barectf_config_parse_common._v3_prefixes_from_v2_prefix(self.cfg.v2_prefix)
            cg_opts = config.options.code_generation_options
            cg_opts = barectf.ConfigurationCodeGenerationOptions(v3_prefixes.identifier,
                                                                 v3_prefixes.file_name,
                                                                 cg_opts.default_data_stream_type,
                                                                 cg_opts.header_options,
                                                                 cg_opts.clock_type_c_types)
            config = barectf.Configuration(config.trace, barectf.ConfigurationOptions(cg_opts))

        # create a barectf code generator
        code_gen = barectf.CodeGenerator(config)

        def write_file(dir, file):
            with open(os.path.join(dir, file.name), 'w') as f:
                f.write(file.contents)

        def write_files(dir, files):
            for file in files:
                write_file(dir, file)

        try:
            # generate and write metadata stream file
            write_file(self.cfg.metadata_stream_dir, code_gen.generate_metadata_stream())

            # generate and write C header files
            write_files(self.cfg.c_header_dir, code_gen.generate_c_headers())

            # generate and write C source files
            write_files(self.cfg.c_source_dir, code_gen.generate_c_sources())
        except Exception as exc:
            # We know `config` is valid, therefore the code generator cannot
            # fail for a reason known to barectf.
            _print_unknown_exc(exc)


class _ShowEffectiveCfgCmdCfg(_CfgCmdCfg):
    def __init__(self, cfg_file_path: str, inclusion_dirs: List[str],
                 ignore_inclusion_file_not_found: bool, indent_space_count: Count):
        super().__init__(cfg_file_path, inclusion_dirs, ignore_inclusion_file_not_found)
        self._indent_space_count = indent_space_count

    @property
    def indent_space_count(self) -> Count:
        return self._indent_space_count


# Effective configuration showing command.
class _ShowEffectiveCfgCmd(_Cmd):
    def exec(self):
        try:
            with open(self.cfg.cfg_file_path) as f:
                print(barectf.effective_configuration_file(f, True, self.cfg.inclusion_dirs,
                                                           self.cfg.ignore_inclusion_file_not_found,
                                                           self.cfg.indent_space_count))
        except barectf._ConfigurationParseError as exc:
            _print_config_error(exc)
        except Exception as exc:
            _print_unknown_exc(exc)


class _ShowCfgVersionCmdCfg(_CmdCfg):
    def __init__(self, cfg_file_path: str):
        self._cfg_file_path = cfg_file_path

    @property
    def cfg_file_path(self) -> str:
        return self._cfg_file_path


class _ShowCfgVersionCmd(_Cmd):
    def exec(self):
        try:
            with open(self.cfg.cfg_file_path) as f:
                print(barectf.configuration_file_major_version(f))
        except barectf._ConfigurationParseError as exc:
            _print_config_error(exc)
        except Exception as exc:
            _print_unknown_exc(exc)


def _run():
    # create command from arguments
    try:
        cmd = _cmd_from_args(sys.argv[1:])
    except barectf_argpar._Error as exc:
        _print_error(f'Command-line: For argument `{exc.orig_arg}`: {exc.msg}')
    except _CliError as exc:
        _print_error(f'Command-line: {exc}')

    # execute command
    cmd.exec()



# The MIT License (MIT)
#
# Copyright (c) 2015-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import barectf.config as barectf_config
import barectf.template as barectf_template
from typing import List, Optional, Union
import typing


def _filt_disp_base_int(disp_base: barectf_config.DisplayBase) -> int:
    return {
        barectf_config.DisplayBase.BINARY: 2,
        barectf_config.DisplayBase.OCTAL: 8,
        barectf_config.DisplayBase.DECIMAL: 10,
        barectf_config.DisplayBase.HEXADECIMAL: 16,
    }[disp_base]


def _filt_int_ft_str(ft: barectf_config._FieldType) -> str:
    return _INT_FT_TEMPL.render(ft=ft,
                                is_signed=isinstance(ft, barectf_config.SignedIntegerFieldType))


def _gen_enum_ft(ft: barectf_config._FieldType) -> str:
    return _ENUM_FT_TEMPL.render(ft=ft)


def _gen_real_ft(ft: barectf_config._FieldType) -> str:
    return _REAL_FT_TEMPL.render(ft=ft)


def _gen_str_ft(ft: barectf_config._FieldType) -> str:
    return _STR_FT_TEMPL.render(ft=ft)


def _filt_ft_lengths(ft: barectf_config._FieldType) -> List[Union[str, int]]:
    lengths: List[Union[str, int]] = []

    while isinstance(ft, barectf_config._ArrayFieldType):
        if type(ft) is barectf_config.StaticArrayFieldType:
            ft = typing.cast(barectf_config.StaticArrayFieldType, ft)
            lengths.append(ft.length)
        else:
            assert type(ft) is barectf_config.DynamicArrayFieldType
            ft = typing.cast(barectf_config.DynamicArrayFieldType, ft)
            lengths.append(typing.cast(str, ft._length_ft_member_name))

        ft = ft.element_field_type

    return lengths


def _filt_deepest_ft(ft: barectf_config._FieldType) -> barectf_config._FieldType:
    while isinstance(ft, barectf_config._ArrayFieldType):
        ft = ft.element_field_type

    return ft


def _gen_struct_ft(ft: barectf_config._FieldType) -> str:
    return _STRUCT_FT_TEMPL.render(ft=ft)


_FT_CLS_TO_GEN_FT_FUNC = {
    barectf_config.UnsignedIntegerFieldType: _filt_int_ft_str,
    barectf_config.SignedIntegerFieldType: _filt_int_ft_str,
    barectf_config.UnsignedEnumerationFieldType: _gen_enum_ft,
    barectf_config.SignedEnumerationFieldType: _gen_enum_ft,
    barectf_config.RealFieldType: _gen_real_ft,
    barectf_config.StringFieldType: _gen_str_ft,
    barectf_config.StructureFieldType: _gen_struct_ft,
}


def _filt_ft_str(ft: barectf_config._FieldType) -> str:
    return _FT_CLS_TO_GEN_FT_FUNC[type(ft)](ft)


_TEMPL_FILTERS = {
    'disp_base_int': _filt_disp_base_int,
    'int_ft_str': _filt_int_ft_str,
    'ft_str': _filt_ft_str,
    'ft_lengths': _filt_ft_lengths,
    'deepest_ft': _filt_deepest_ft,
}


def _create_template(name: str, is_file_template: bool = False,
                     cfg: Optional[barectf_config.Configuration] = None) -> barectf_template._Template:
    return barectf_template._Template(f'metadata/{name}', is_file_template, cfg,
                                      typing.cast(barectf_template._Filters, _TEMPL_FILTERS))


_ENUM_FT_TEMPL = _create_template('enum-ft.j2')
_INT_FT_TEMPL = _create_template('int-ft.j2')
_REAL_FT_TEMPL = _create_template('real-ft.j2')
_STR_FT_TEMPL = _create_template('str-ft.j2')
_STRUCT_FT_TEMPL = _create_template('struct-ft.j2')


def _from_config(cfg: barectf_config.Configuration) -> str:
    return _create_template('metadata.j2', True, cfg).render()



# The MIT License (MIT)
#
# Copyright (c) 2015-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documeneffective_filetation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import pkg_resources
import collections
import jsonschema  # type: ignore
import os.path
import yaml
import copy
import os
from barectf.typing import VersionNumber, _OptStr
from typing import Optional, List, Dict, Any, TextIO, MutableMapping, Union, Set, Iterable, Callable, Tuple
import typing


# The context of a configuration parsing error.
#
# Such a context object has a name and, optionally, a message.
class _ConfigurationParseErrorContext:
    def __init__(self, name: str, message: _OptStr = None):
        self._name = name
        self._msg = message

    @property
    def name(self) -> str:
        return self._name

    @property
    def message(self) -> _OptStr:
        return self._msg


# A configuration parsing error.
#
# Such an error object contains a list of contexts (`context` property).
#
# The first context of this list is the most specific context, while the
# last is the more general.
#
# Use _append_ctx() to append a context to an existing configuration
# parsing error when you catch it before raising it again. You can use
# _append_error_ctx() to do exactly this in a single call.
class _ConfigurationParseError(Exception):
    def __init__(self, init_ctx_obj_name, init_ctx_msg=None):
        super().__init__()
        self._ctx: List[_ConfigurationParseErrorContext] = []
        self._append_ctx(init_ctx_obj_name, init_ctx_msg)

    @property
    def context(self) -> List[_ConfigurationParseErrorContext]:
        return self._ctx

    def _append_ctx(self, name: str, msg: _OptStr = None):
        self._ctx.append(_ConfigurationParseErrorContext(name, msg))

    def __str__(self):
        lines = []

        for ctx in reversed(self._ctx):
            line = f'{ctx.name}:'

            if ctx.message is not None:
                line += f' {ctx.message}'

            lines.append(line)

        return '\n'.join(lines)


# Appends the context having the object name `obj_name` and the
# (optional) message `message` to the `_ConfigurationParseError`
# exception `exc` and then raises `exc` again.
def _append_error_ctx(exc: _ConfigurationParseError, obj_name: str, message: _OptStr = None):
    exc._append_ctx(obj_name, message)
    raise exc


_V3Prefixes = collections.namedtuple('_V3Prefixes', ['identifier', 'file_name'])


# Convers a v2 prefix to v3 prefixes.
def _v3_prefixes_from_v2_prefix(v2_prefix: str) -> _V3Prefixes:
    return _V3Prefixes(v2_prefix, v2_prefix.rstrip('_'))


# This JSON schema reference resolver only serves to detect when it
# needs to resolve a remote URI.
#
# This must never happen in barectf because all our schemas are local;
# it would mean a programming or schema error.
class _RefResolver(jsonschema.RefResolver):
    def resolve_remote(self, uri: str):
        raise RuntimeError(f'Missing local schema with URI `{uri}`')


# Not all static type checkers support type recursion, so let's just use
# `Any` as a map node's value's type.
_MapNode = MutableMapping[str, Any]


# Schema validator which considers all the schemas found in the
# subdirectories `subdirs` (at build time) of the barectf package's
# `schemas` directory.
#
# The only public method is validate() which accepts an instance to
# validate as well as a schema short ID.
class _SchemaValidator:
    def __init__(self, subdirs: Iterable[str]):
        schemas_dir = pkg_resources.resource_filename(__name__, 'schemas')
        self._store: Dict[str, str] = {}

        for subdir in subdirs:
            dir = os.path.join(schemas_dir, subdir)

            for file_name in os.listdir(dir):
                if not file_name.endswith('.yaml'):
                    continue

                with open(os.path.join(dir, file_name)) as f:
                    schema = yaml.load(f, Loader=yaml.SafeLoader)

                assert '$id' in schema
                schema_id = schema['$id']
                assert schema_id not in self._store
                self._store[schema_id] = schema

    @staticmethod
    def _dict_from_ordered_dict(obj):
        if type(obj) is not collections.OrderedDict:
            return obj

        dct = {}

        for k, v in obj.items():
            new_v = v

            if type(v) is collections.OrderedDict:
                new_v = _SchemaValidator._dict_from_ordered_dict(v)
            elif type(v) is list:
                new_v = [_SchemaValidator._dict_from_ordered_dict(elem) for elem in v]

            dct[k] = new_v

        return dct

    def _validate(self, instance: _MapNode, schema_short_id: str):
        # retrieve full schema ID from short ID
        schema_id = f'https://barectf.org/schemas/{schema_short_id}.json'
        assert schema_id in self._store

        # retrieve full schema
        schema = self._store[schema_id]

        # Create a reference resolver for this schema using this
        # validator's schema store.
        resolver = _RefResolver(base_uri=schema_id, referrer=schema,
                                store=self._store)

        # create a JSON schema validator using this reference resolver
        validator = jsonschema.Draft7Validator(schema, resolver=resolver)

        # Validate the instance, converting its
        # `collections.OrderedDict` objects to `dict` objects so as to
        # make any error message easier to read (because
        # validator.validate() below uses str() for error messages, and
        # collections.OrderedDict.__str__() returns a somewhat bulky
        # representation).
        validator.validate(self._dict_from_ordered_dict(instance))

    # Validates `instance` using the schema having the short ID
    # `schema_short_id`.
    #
    # A schema short ID is the part between `schemas/` and `.json` in
    # its URI.
    #
    # Raises a `_ConfigurationParseError` object, hiding any
    # `jsonschema` exception, on validation failure.
    def validate(self, instance: _MapNode, schema_short_id: str):
        try:
            self._validate(instance, schema_short_id)
        except jsonschema.ValidationError as exc:
            # convert to barectf `_ConfigurationParseError` exception
            contexts = ['Configuration object']

            # Each element of the instance's absolute path is either an
            # integer (array element's index) or a string (object
            # property's name).
            for elem in exc.absolute_path:
                if type(elem) is int:
                    ctx = f'Element #{elem + 1}'
                else:
                    ctx = f'`{elem}` property'

                contexts.append(ctx)

            schema_ctx = ''

            if len(exc.context) > 0:
                # According to the documentation of
                # jsonschema.ValidationError.context(), the method
                # returns a
                #
                # > list of errors from the subschemas
                #
                # This contains additional information about the
                # validation failure which can help the user figure out
                # what's wrong exactly.
                #
                # Join each message with `; ` and append this to our
                # configuration parsing error's message.
                msgs = '; '.join([e.message for e in exc.context])
                schema_ctx = f': {msgs}'

            new_exc = _ConfigurationParseError(contexts.pop(),
                                               f'{exc.message}{schema_ctx} (from schema `{schema_short_id}`)')

            for ctx in reversed(contexts):
                new_exc._append_ctx(ctx)

            raise new_exc


# barectf 3 YAML configuration node.
class _ConfigNodeV3:
    def __init__(self, config_node: _MapNode):
        self._config_node = config_node

    @property
    def config_node(self) -> _MapNode:
        return self._config_node


_CONFIG_V3_YAML_TAG = 'tag:barectf.org,2020/3/config'


# Loads the content of the YAML file-like object `file` as a Python
# object and returns it.
#
# If the file's object has the barectf 3 configuration tag, then this
# function returns a `_ConfigNodeV3` object. Otherwise, it returns a
# `collections.OrderedDict` object.
#
# All YAML maps are loaded as `collections.OrderedDict` objects.
def _yaml_load(file: TextIO) -> Union[_ConfigNodeV3, _MapNode]:
    class Loader(yaml.Loader):
        pass

    def config_ctor(loader, node) -> _ConfigNodeV3:
        if not isinstance(node, yaml.MappingNode):
            problem = f'Expecting a map for the tag `{node.tag}`'
            raise yaml.constructor.ConstructorError(problem=problem)

        loader.flatten_mapping(node)
        return _ConfigNodeV3(collections.OrderedDict(loader.construct_pairs(node)))

    def mapping_ctor(loader, node) -> _MapNode:
        loader.flatten_mapping(node)
        return collections.OrderedDict(loader.construct_pairs(node))

    Loader.add_constructor(_CONFIG_V3_YAML_TAG, config_ctor)
    Loader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, mapping_ctor)

    # YAML -> Python
    try:
        return yaml.load(file, Loader=Loader)
    except (yaml.YAMLError, OSError, IOError) as exc:
        raise _ConfigurationParseError('YAML loader', f'Cannot load file: {exc}')


def _yaml_load_path(path: str) -> Union[_ConfigNodeV3, _MapNode]:
    with open(path) as f:
        return _yaml_load(f)


# Dumps the content of the Python object `obj`
# (`collections.OrderedDict` or `_ConfigNodeV3`) as a YAML string and
# returns it.
def _yaml_dump(node: _MapNode, **kwds) -> str:
    class Dumper(yaml.Dumper):
        pass

    def config_repr(dumper, node):
        return dumper.represent_mapping(_CONFIG_V3_YAML_TAG, node.config_node.items())

    def mapping_repr(dumper, node):
        return dumper.represent_mapping(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                                        node.items())

    Dumper.add_representer(_ConfigNodeV3, config_repr)
    Dumper.add_representer(collections.OrderedDict, mapping_repr)

    # Python -> YAML
    return yaml.dump(node, Dumper=Dumper, version=(1, 2), **kwds)


# A common barectf YAML configuration parser.
#
# This is the base class of any barectf YAML configuration parser. It
# mostly contains helpers.
class _Parser:
    # Builds a base barectf YAML configuration parser to process the
    # configuration node `node` (already loaded from the file-like
    # object `file`).
    #
    # For its _process_node_include() method, the parser considers the
    # package inclusion directory as well as `include_dirs`, and ignores
    # nonexistent inclusion files if `ignore_include_not_found` is
    # `True`.
    def __init__(self, root_file: TextIO, node: Union[_MapNode, _ConfigNodeV3],
                 with_pkg_include_dir: bool, include_dirs: Optional[List[str]],
                 ignore_include_not_found: bool, major_version: VersionNumber):
        self._root_file = root_file
        self._root_node = node
        self._ft_prop_names = [
            # barectf 2.1+
            '$inherit',

            # barectf 2
            'inherit',
            'value-type',
            'element-type',

            # barectf 3
            'element-field-type',
        ]

        if include_dirs is None:
            include_dirs = []

        self._include_dirs = copy.copy(include_dirs)

        if with_pkg_include_dir:
            self._include_dirs.append(pkg_resources.resource_filename(__name__, f'include/{major_version}'))

        self._ignore_include_not_found = ignore_include_not_found
        self._include_stack: List[str] = []
        self._resolved_ft_aliases: Set[str] = set()
        self._schema_validator = _SchemaValidator({'config/common', f'config/{major_version}'})
        self._major_version = major_version

    @property
    def root_node(self):
        return self._root_node

    @property
    def _struct_ft_node_members_prop_name(self) -> str:
        if self._major_version == 2:
            return 'fields'
        else:
            return 'members'

    # Returns the last included file name from the parser's inclusion
    # file name stack, or `N/A` if the root file does not have an
    # associated path under the `name` property.
    def _get_last_include_file(self) -> str:
        if self._include_stack:
            return self._include_stack[-1]

        if hasattr(self._root_file, 'name'):
            return typing.cast(str, self._root_file.name)

        return 'N/A'

    # Loads the inclusion file having the path `yaml_path` and returns
    # its content as a `collections.OrderedDict` object.
    def _load_include(self, yaml_path) -> Optional[_MapNode]:
        for inc_dir in self._include_dirs:
            # Current inclusion dir + file name path.
            #
            # Note: os.path.join() only takes the last argument if it's
            # absolute.
            inc_path = os.path.join(inc_dir, yaml_path)

            # real path (symbolic links resolved)
            real_path = os.path.realpath(inc_path)

            # normalized path (weird stuff removed!)
            norm_path = os.path.normpath(real_path)

            if not os.path.isfile(norm_path):
                # file doesn't exist: skip
                continue

            if norm_path in self._include_stack:
                base_path = self._get_last_include_file()
                raise _ConfigurationParseError(f'File `{base_path}`',
                                               f'Cannot recursively include file `{norm_path}`')

            self._include_stack.append(norm_path)

            # load raw content
            return typing.cast(_MapNode, _yaml_load_path(norm_path))

        if not self._ignore_include_not_found:
            base_path = self._get_last_include_file()
            raise _ConfigurationParseError(f'File `{base_path}`',
                                           f'Cannot include file `{yaml_path}`: file not found in inclusion directories')

        return None

    # Returns a list of all the inclusion file paths as found in the
    # inclusion node `include_node`.
    def _get_include_paths(self, include_node: _MapNode) -> List[str]:
        if include_node is None:
            # none
            return []

        if type(include_node) is str:
            # wrap as array
            return [typing.cast(str, include_node)]

        # already an array
        assert type(include_node) is list
        return typing.cast(List[str], include_node)

    # Updates the node `base_node` with an overlay node `overlay_node`.
    #
    # Both the inclusion and field type node inheritance features use
    # this update mechanism.
    def _update_node(self, base_node: _MapNode, overlay_node: _MapNode):
        # see the comment about the `members` property below
        def update_members_node(base_value: List[Any], olay_value: List[Any]):
            for olay_item in olay_value:
                # assume we append `olay_item` to `base_value` initially
                append_olay_item = True

                if type(olay_item) is collections.OrderedDict:
                    # overlay item is an object
                    if len(olay_item) == 1:
                        # overlay object item contains a single property
                        olay_name = list(olay_item)[0]

                        # find corresponding base item
                        for base_item in base_value:
                            if type(base_item) is collections.OrderedDict:
                                if len(olay_item) == 1:
                                    base_name = list(base_item)[0]

                                    if olay_name == base_name:
                                        # Names match: update with usual
                                        # strategy.
                                        self._update_node(base_item, olay_item)

                                        # Do _not_ append `olay_item` to
                                        # `base_value`: we just updated
                                        # `base_item`.
                                        append_olay_item = False
                                        break

                if append_olay_item:
                    base_value.append(copy.deepcopy(olay_item))

        for olay_key, olay_value in overlay_node.items():
            if olay_key in base_node:
                base_value = base_node[olay_key]

                if type(olay_value) is collections.OrderedDict and type(base_value) is collections.OrderedDict:
                    # merge both objects
                    self._update_node(base_value, olay_value)
                elif type(olay_value) is list and type(base_value) is list:
                    if olay_key == 'members' and self._major_version == 3:
                        # This is a "temporary" hack.
                        #
                        # In barectf 2, a structure field type node
                        # looks like this:
                        #
                        #     class: struct
                        #     fields:
                        #       hello: uint8
                        #       world: string
                        #
                        # Having an overlay such as
                        #
                        #     fields:
                        #       hello: float
                        #
                        # will result in
                        #
                        #     class: struct
                        #     fields:
                        #       hello: float
                        #       world: string
                        #
                        # because the `fields` property is a map.
                        #
                        # In barectf 3, this is fixed (a YAML map is not
                        # ordered), so that the same initial structure
                        # field type node looks like this:
                        #
                        #     class: struct
                        #     members:
                        #       - hello: uint8
                        #       - world:
                        #           field-type:
                        #             class: str
                        #
                        # Although the `members` property is
                        # syntaxically an array, it's semantically an
                        # ordered map, where an entry's key is the array
                        # item's map's first key (like YAML's `!!omap`).
                        #
                        # Having an overlay such as
                        #
                        #     members:
                        #       - hello: float
                        #
                        # would result in
                        #
                        #     class: struct
                        #     members:
                        #       - hello: uint8
                        #       - world:
                        #           field-type:
                        #             class: str
                        #       - hello: float
                        #
                        # with the naive strategy, while what we really
                        # want is:
                        #
                        #     class: struct
                        #     members:
                        #       - hello: float
                        #       - world:
                        #           field-type:
                        #             class: str
                        #
                        # As of this version of barectf, the _only_
                        # property with a list value which acts as an
                        # ordered map is named `members`. This is why we
                        # can only check the value of `olay_key`,
                        # whatever our context.
                        #
                        # update_members_node() attempts to perform
                        # this below. For a given item of `olay_value`,
                        # if
                        #
                        # * It's not an object.
                        #
                        # * It contains more than one property.
                        #
                        # * Its single property's name does not match
                        #   the name of the single property of any
                        #   object item of `base_value`.
                        #
                        # then we append the item to `base_value` as
                        # usual.
                        update_members_node(base_value, olay_value)
                    else:
                        # append extension array items to base items
                        base_value += copy.deepcopy(olay_value)
                else:
                    # fall back to replacing base property
                    base_node[olay_key] = copy.deepcopy(olay_value)
            else:
                # set base property from overlay property
                base_node[olay_key] = copy.deepcopy(olay_value)

    # Processes inclusions using `last_overlay_node` as the last overlay
    # node to use to "patch" the node.
    #
    # If `last_overlay_node` contains an `$include` property, then this
    # method patches the current base node (initially empty) in order
    # using the content of the inclusion files (recursively).
    #
    # At the end, this method removes the `$include` property of
    # `last_overlay_node` and then patches the current base node with
    # its other properties before returning the result (always a deep
    # copy).
    def _process_node_include(self, last_overlay_node: _MapNode,
                              process_base_include_cb: Callable[[_MapNode], _MapNode],
                              process_children_include_cb: Optional[Callable[[_MapNode], None]] = None) -> _MapNode:
        # process children inclusions first
        if process_children_include_cb is not None:
            process_children_include_cb(last_overlay_node)

        incl_prop_name = '$include'

        if incl_prop_name in last_overlay_node:
            include_node = last_overlay_node[incl_prop_name]
        else:
            # no inclusions!
            return last_overlay_node

        include_paths = self._get_include_paths(include_node)
        cur_base_path = self._get_last_include_file()
        base_node = None

        # keep the inclusion paths and remove the `$include` property
        include_paths = copy.deepcopy(include_paths)
        del last_overlay_node[incl_prop_name]

        for include_path in include_paths:
            # load raw YAML from included file
            overlay_node = self._load_include(include_path)

            if overlay_node is None:
                # Cannot find inclusion file, but we're ignoring those
                # errors, otherwise _load_include() itself raises a
                # config error.
                continue

            # recursively process inclusions
            try:
                overlay_node = process_base_include_cb(overlay_node)
            except _ConfigurationParseError as exc:
                _append_error_ctx(exc, f'File `{cur_base_path}`')

            # pop inclusion stack now that we're done including
            del self._include_stack[-1]

            # At this point, `base_node` is fully resolved (does not
            # contain any `$include` property).
            if base_node is None:
                base_node = overlay_node
            else:
                self._update_node(base_node, overlay_node)

        # Finally, update the latest base node with our last overlay
        # node.
        if base_node is None:
            # Nothing was included, which is possible when we're
            # ignoring inclusion errors.
            return last_overlay_node

        self._update_node(base_node, last_overlay_node)
        return base_node

    # Generates pairs of member node and field type node property name
    # (in the member node) for the structure field type node's members
    # node `node`.
    def _struct_ft_member_fts_iter(self,
                                   node: Union[List[_MapNode], _MapNode]) -> Iterable[Tuple[_MapNode, str]]:
        if type(node) is list:
            # barectf 3
            assert self._major_version == 3
            node = typing.cast(List[_MapNode], node)

            for member_node in node:
                assert type(member_node) is collections.OrderedDict
                member_node = typing.cast(_MapNode, member_node)
                name, val = list(member_node.items())[0]

                if type(val) is collections.OrderedDict:
                    member_node = val
                    name = 'field-type'

                yield member_node, name
        else:
            # barectf 2
            assert self._major_version == 2
            assert type(node) is collections.OrderedDict
            node = typing.cast(_MapNode, node)

            for name in node:
                yield node, name

    # Resolves the field type alias `key` in the node `parent_node`, as
    # well as any nested field type aliases, using the aliases of the
    # `ft_aliases_node` node.
    #
    # If `key` is not in `parent_node`, this method returns.
    #
    # This method can modify `ft_aliases_node` and `parent_node[key]`.
    #
    # `ctx_obj_name` is the context's object name when this method
    # raises a `_ConfigurationParseError` exception.
    def _resolve_ft_alias(self, ft_aliases_node: _MapNode, parent_node: _MapNode, key: str,
                          ctx_obj_name: str, alias_set: Optional[Set[str]] = None):
        if key not in parent_node:
            return

        node = parent_node[key]

        if node is None:
            # some nodes can be null to use their default value
            return

        # This set holds all the field type aliases to be expanded,
        # recursively. This is used to detect cycles.
        if alias_set is None:
            alias_set = set()

        if type(node) is str:
            alias = node

            # Make sure this alias names an existing field type node, at
            # least.
            if alias not in ft_aliases_node:
                raise _ConfigurationParseError(ctx_obj_name,
                                               f'Field type alias `{alias}` does not exist')

            if alias not in self._resolved_ft_aliases:
                # Only check for a field type alias cycle when we didn't
                # resolve the alias yet, as a given node can refer to
                # the same field type alias more than once.
                if alias in alias_set:
                    msg = f'Cycle detected during the `{alias}` field type alias resolution'
                    raise _ConfigurationParseError(ctx_obj_name, msg)

                # Resolve it.
                #
                # Add `alias` to the set of encountered field type
                # aliases before calling self._resolve_ft_alias() to
                # detect cycles.
                alias_set.add(alias)
                self._resolve_ft_alias(ft_aliases_node, ft_aliases_node, alias, ctx_obj_name,
                                       alias_set)
                self._resolved_ft_aliases.add(alias)

            # replace alias with field type node copy
            parent_node[key] = copy.deepcopy(ft_aliases_node[alias])
            return

        # resolve nested field type aliases
        for pkey in self._ft_prop_names:
            self._resolve_ft_alias(ft_aliases_node, node, pkey, ctx_obj_name, alias_set)

        # Resolve field type aliases of structure field type node member
        # nodes.
        pkey = self._struct_ft_node_members_prop_name

        if pkey in node:
            for member_node, ft_prop_name in self._struct_ft_member_fts_iter(node[pkey]):
                self._resolve_ft_alias(ft_aliases_node, member_node, ft_prop_name,
                                       ctx_obj_name, alias_set)

    # Like _resolve_ft_alias(), but builds a context object name for any
    # `ctx_obj_name` exception.
    def _resolve_ft_alias_from(self, ft_aliases_node: _MapNode, parent_node: _MapNode, key: str):
        self._resolve_ft_alias(ft_aliases_node, parent_node, key, f'`{key}` property')

    # Applies field type node inheritance to the property `key` of
    # `parent_node`.
    #
    # `parent_node[key]`, if it exists, must not contain any field type
    # alias (all field type objects are complete).
    #
    # This method can modify `parent[key]`.
    #
    # When this method returns, no field type node has an `$inherit` or
    # `inherit` property.
    def _apply_ft_inheritance(self, parent_node: _MapNode, key: str):
        if key not in parent_node:
            return

        node = parent_node[key]

        if node is None:
            return

        # process children first
        for pkey in self._ft_prop_names:
            self._apply_ft_inheritance(node, pkey)

        # Process the field types of structure field type node member
        # nodes.
        pkey = self._struct_ft_node_members_prop_name

        if pkey in node:
            for member_node, ft_prop_name in self._struct_ft_member_fts_iter(node[pkey]):
                self._apply_ft_inheritance(member_node, ft_prop_name)

        # apply inheritance for this node
        if 'inherit' in node:
            # barectf 2.1: `inherit` property was renamed to `$inherit`
            assert '$inherit' not in node
            node['$inherit'] = node['inherit']
            del node['inherit']

        inherit_key = '$inherit'

        if inherit_key in node:
            assert type(node[inherit_key]) is collections.OrderedDict

            # apply inheritance below
            self._apply_ft_inheritance(node, inherit_key)

            # `node` is an overlay on the `$inherit` node
            base_node = node[inherit_key]
            del node[inherit_key]
            self._update_node(base_node, node)

            # set updated base node as this node
            parent_node[key] = base_node



# The MIT License (MIT)
#
# Copyright (c) 2015-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import barectf.version as barectf_version
from typing import Optional, Any, FrozenSet, Mapping, Iterator, Set, Union, Callable
import typing
from barectf.typing import Count, Alignment, _OptStr, Id
import collections.abc
import collections
import datetime
import enum
import uuid as uuidp


@enum.unique
class ByteOrder(enum.Enum):
    LITTLE_ENDIAN = 'le'
    BIG_ENDIAN = 'be'


class _FieldType:
    @property
    def alignment(self) -> Alignment:
        raise NotImplementedError

    @property
    def size_is_dynamic(self):
        return False


class _BitArrayFieldType(_FieldType):
    def __init__(self, size: Count, alignment: Alignment = Alignment(1)):
        self._size = size
        self._alignment = alignment

    @property
    def size(self) -> Count:
        return self._size

    @property
    def alignment(self) -> Alignment:
        return self._alignment


class DisplayBase(enum.Enum):
    BINARY = 2
    OCTAL = 8
    DECIMAL = 10
    HEXADECIMAL = 16


class _IntegerFieldType(_BitArrayFieldType):
    def __init__(self, size: Count, alignment: Optional[Alignment] = None,
                 preferred_display_base: DisplayBase = DisplayBase.DECIMAL):
        if alignment is None:
            alignment = Alignment(8 if size % 8 == 0 else 1)

        super().__init__(size, alignment)
        self._preferred_display_base = preferred_display_base

    @property
    def preferred_display_base(self) -> DisplayBase:
        return self._preferred_display_base


class UnsignedIntegerFieldType(_IntegerFieldType):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mapped_clk_type_name = None
        self._is_len = False


class SignedIntegerFieldType(_IntegerFieldType):
    pass


class EnumerationFieldTypeMappingRange:
    def __init__(self, lower: int, upper: int):
        self._lower = lower
        self._upper = upper

    @property
    def lower(self) -> int:
        return self._lower

    @property
    def upper(self) -> int:
        return self._upper

    def __eq__(self, other: Any) -> bool:
        if type(other) is not type(self):
            return False

        return (self._lower, self._upper) == (other._lower, other._upper)

    def __hash__(self) -> int:
        return hash((self._lower, self._upper))

    def contains(self, value: int) -> bool:
        return self._lower <= value <= self._upper


class EnumerationFieldTypeMapping:
    def __init__(self, ranges: Set[EnumerationFieldTypeMappingRange]):
        self._ranges = frozenset(ranges)

    @property
    def ranges(self) -> FrozenSet[EnumerationFieldTypeMappingRange]:
        return self._ranges

    def ranges_contain_value(self, value: int) -> bool:
        return any([rg.contains(value) for rg in self._ranges])


_EnumFtMappings = Mapping[str, EnumerationFieldTypeMapping]


class EnumerationFieldTypeMappings(collections.abc.Mapping):
    def __init__(self, mappings: _EnumFtMappings):
        self._mappings = {label: mapping for label, mapping in mappings.items()}

    def __getitem__(self, key: str) -> EnumerationFieldTypeMapping:
        return self._mappings[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._mappings)

    def __len__(self) -> int:
        return len(self._mappings)


class _EnumerationFieldType(_IntegerFieldType):
    def __init__(self, size: Count, alignment: Optional[Alignment] = None,
                 preferred_display_base: DisplayBase = DisplayBase.DECIMAL,
                 mappings: Optional[_EnumFtMappings] = None):
        super().__init__(size, alignment, preferred_display_base)
        self._mappings = EnumerationFieldTypeMappings({})

        if mappings is not None:
            self._mappings = EnumerationFieldTypeMappings(mappings)

    @property
    def mappings(self) -> EnumerationFieldTypeMappings:
        return self._mappings

    def labels_for_value(self, value: int) -> Set[str]:
        labels = set()

        for label, mapping in self._mappings.items():
            if mapping.ranges_contain_value(value):
                labels.add(label)

        return labels


class UnsignedEnumerationFieldType(_EnumerationFieldType, UnsignedIntegerFieldType):
    pass


class SignedEnumerationFieldType(_EnumerationFieldType, SignedIntegerFieldType):
    pass


class RealFieldType(_BitArrayFieldType):
    pass


class StringFieldType(_FieldType):
    @property
    def alignment(self) -> Alignment:
        return Alignment(8)

    @property
    def size_is_dynamic(self):
        return True


class _ArrayFieldType(_FieldType):
    def __init__(self, element_field_type: _FieldType):
        self._element_field_type = element_field_type

    @property
    def element_field_type(self) -> _FieldType:
        return self._element_field_type

    @property
    def alignment(self) -> Alignment:
        return self._element_field_type.alignment

    @property
    def size_is_dynamic(self):
        return self._element_field_type.size_is_dynamic


class StaticArrayFieldType(_ArrayFieldType):
    def __init__(self, length: Count, element_field_type: _FieldType):
        super().__init__(element_field_type)
        self._length = length

    @property
    def length(self) -> Count:
        return self._length


class DynamicArrayFieldType(_ArrayFieldType):
    def __init__(self, length_field_type: UnsignedIntegerFieldType, element_field_type: _FieldType):
        super().__init__(element_field_type)
        self._length_field_type = length_field_type
        self._length_ft_member_name: Optional[str] = None

    @property
    def length_field_type(self):
        return self._length_field_type

    @property
    def size_is_dynamic(self):
        return True


class StructureFieldTypeMember:
    def __init__(self, field_type: _FieldType):
        self._field_type = field_type

    @property
    def field_type(self) -> _FieldType:
        return self._field_type


_StructFtMembers = Mapping[str, StructureFieldTypeMember]


class StructureFieldTypeMembers(collections.abc.Mapping):
    def __init__(self, members: _StructFtMembers):
        self._members = collections.OrderedDict()

        for name, member in members.items():
            assert type(member) is StructureFieldTypeMember
            self._members[name] = member

    def __getitem__(self, key: str) -> StructureFieldTypeMember:
        return self._members[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._members)

    def __len__(self) -> int:
        return len(self._members)


class StructureFieldType(_FieldType):
    def __init__(self, minimum_alignment: Alignment = Alignment(1),
                 members: Optional[_StructFtMembers] = None):
        self._minimum_alignment = minimum_alignment
        self._members = StructureFieldTypeMembers({})

        if members is not None:
            self._members = StructureFieldTypeMembers(members)

        self._set_alignment()
        self._set_dyn_array_ft_length_ft_member_names()

    def _set_alignment(self):
        self._alignment: Alignment = self._minimum_alignment

        for member in self._members.values():
            if member.field_type.alignment > self._alignment:
                self._alignment = member.field_type.alignment

    def _set_dyn_array_ft_length_ft_member_names(self):
        for member in self._members.values():
            if type(member.field_type) is DynamicArrayFieldType:
                # Find length field type member name within the same
                # structure field type members.
                for len_name, len_member in self._members.items():
                    if member.field_type.length_field_type is len_member.field_type:
                        member.field_type._length_ft_member_name = len_name
                        len_member.field_type._is_len = True
                        break

            if member.field_type.alignment > self._alignment:
                self._alignment = member.field_type.alignment

    @property
    def minimum_alignment(self) -> Alignment:
        return self._minimum_alignment

    @property
    def alignment(self) -> Alignment:
        return self._alignment

    @property
    def size_is_dynamic(self):
        return any([member.field_type.size_is_dynamic for member in self.members.values()])

    @property
    def members(self) -> StructureFieldTypeMembers:
        return self._members


class _UniqueByName:
    _name: str

    def __eq__(self, other: Any) -> bool:
        if type(other) is not type(self):
            return False

        return self._name == other._name

    def __lt__(self, other: '_UniqueByName'):
        assert type(self) is type(other)
        return self._name < other._name

    def __hash__(self) -> int:
        return hash(self._name)


_OptFt = Optional[_FieldType]
_OptStructFt = Optional[StructureFieldType]
LogLevel = typing.NewType('LogLevel', int)


class EventRecordType(_UniqueByName):
    def __init__(self, name: str, log_level: Optional[LogLevel] = None,
                 specific_context_field_type: _OptStructFt = None, payload_field_type: _OptStructFt = None):
        self._id: Optional[Id] = None
        self._name = name
        self._log_level = log_level
        self._specific_context_field_type = specific_context_field_type
        self._payload_field_type = payload_field_type

    @property
    def id(self) -> Optional[Id]:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def log_level(self) -> Optional[LogLevel]:
        return self._log_level

    @property
    def specific_context_field_type(self) -> _OptStructFt:
        return self._specific_context_field_type

    @property
    def payload_field_type(self) -> _OptStructFt:
        return self._payload_field_type


class ClockTypeOffset:
    def __init__(self, seconds: int = 0, cycles: Count = Count(0)):
        self._seconds = seconds
        self._cycles = cycles

    @property
    def seconds(self) -> int:
        return self._seconds

    @property
    def cycles(self) -> Count:
        return self._cycles


_OptUuid = Optional[uuidp.UUID]


class ClockType(_UniqueByName):
    def __init__(self, name: str, frequency: Count = Count(int(1e9)), uuid: _OptUuid = None,
                 description: _OptStr = None, precision: Count = Count(0),
                 offset: Optional[ClockTypeOffset] = None, origin_is_unix_epoch: bool = False):
        self._name = name
        self._frequency = frequency
        self._uuid = uuid
        self._description = description
        self._precision = precision
        self._offset = ClockTypeOffset()

        if offset is not None:
            self._offset = offset

        self._origin_is_unix_epoch = origin_is_unix_epoch

    @property
    def name(self) -> str:
        return self._name

    @property
    def frequency(self) -> Count:
        return self._frequency

    @property
    def uuid(self) -> _OptUuid:
        return self._uuid

    @property
    def description(self) -> _OptStr:
        return self._description

    @property
    def precision(self) -> Count:
        return self._precision

    @property
    def offset(self) -> ClockTypeOffset:
        return self._offset

    @property
    def origin_is_unix_epoch(self) -> bool:
        return self._origin_is_unix_epoch


DEFAULT_FIELD_TYPE = 'default'
_DefaultableUIntFt = Union[str, UnsignedIntegerFieldType]
_OptDefaultableUIntFt = Optional[_DefaultableUIntFt]
_OptUIntFt = Optional[UnsignedIntegerFieldType]


class DataStreamTypePacketFeatures:
    def __init__(self, total_size_field_type: _DefaultableUIntFt = DEFAULT_FIELD_TYPE,
                 content_size_field_type: _DefaultableUIntFt = DEFAULT_FIELD_TYPE,
                 beginning_timestamp_field_type: _OptDefaultableUIntFt = None,
                 end_timestamp_field_type: _OptDefaultableUIntFt = None,
                 discarded_event_records_snapshot_counter_field_type: _OptDefaultableUIntFt = DEFAULT_FIELD_TYPE,
                 sequence_number_field_type: _OptDefaultableUIntFt = None):
        def get_ft(user_ft: _OptDefaultableUIntFt) -> _OptUIntFt:
            if user_ft == DEFAULT_FIELD_TYPE:
                return UnsignedIntegerFieldType(64)

            return typing.cast(_OptUIntFt, user_ft)

        self._total_size_field_type = get_ft(total_size_field_type)
        self._content_size_field_type = get_ft(content_size_field_type)
        self._beginning_timestamp_field_type = get_ft(beginning_timestamp_field_type)
        self._end_timestamp_field_type = get_ft(end_timestamp_field_type)
        self._discarded_event_records_snapshot_counter_field_type = get_ft(discarded_event_records_snapshot_counter_field_type)
        self._sequence_number_field_type = get_ft(sequence_number_field_type)

    @property
    def sequence_number_field_type(self) -> _OptUIntFt:
        return self._sequence_number_field_type

    @property
    def total_size_field_type(self) -> _OptUIntFt:
        return self._total_size_field_type

    @property
    def content_size_field_type(self) -> _OptUIntFt:
        return self._content_size_field_type

    @property
    def beginning_timestamp_field_type(self) -> _OptUIntFt:
        return self._beginning_timestamp_field_type

    @property
    def end_timestamp_field_type(self) -> _OptUIntFt:
        return self._end_timestamp_field_type

    @property
    def discarded_event_records_snapshot_counter_field_type(self) -> _OptUIntFt:
        return self._discarded_event_records_snapshot_counter_field_type


class DataStreamTypeEventRecordFeatures:
    def __init__(self, type_id_field_type: _OptDefaultableUIntFt = DEFAULT_FIELD_TYPE,
                 timestamp_field_type: _OptDefaultableUIntFt = None):
        def get_ft(user_ft: _OptDefaultableUIntFt) -> _OptUIntFt:
            if user_ft == DEFAULT_FIELD_TYPE:
                return UnsignedIntegerFieldType(64)

            return typing.cast(_OptUIntFt, user_ft)

        self._type_id_field_type = get_ft(type_id_field_type)
        self._timestamp_field_type = get_ft(timestamp_field_type)

    @property
    def type_id_field_type(self) -> _OptUIntFt:
        return self._type_id_field_type

    @property
    def timestamp_field_type(self) -> _OptUIntFt:
        return self._timestamp_field_type


class DataStreamTypeFeatures:
    def __init__(self, packet_features: Optional[DataStreamTypePacketFeatures] = None,
                 event_record_features: Optional[DataStreamTypeEventRecordFeatures] = None):
        if packet_features is None:
            self._packet_features = DataStreamTypePacketFeatures()
        else:
            self._packet_features = packet_features

        if event_record_features is None:
            self._event_record_features = DataStreamTypeEventRecordFeatures()
        else:
            self._event_record_features = event_record_features

    @property
    def packet_features(self) -> DataStreamTypePacketFeatures:
        return self._packet_features

    @property
    def event_record_features(self) -> DataStreamTypeEventRecordFeatures:
        return self._event_record_features


class DataStreamType(_UniqueByName):
    def __init__(self, name: str, event_record_types: Set[EventRecordType],
                 default_clock_type: Optional[ClockType] = None,
                 features: Optional[DataStreamTypeFeatures] = None,
                 packet_context_field_type_extra_members: Optional[_StructFtMembers] = None,
                 event_record_common_context_field_type: _OptStructFt = None):
        self._id: Optional[Id] = None
        self._name = name
        self._default_clock_type = default_clock_type
        self._event_record_common_context_field_type = event_record_common_context_field_type
        self._event_record_types = frozenset(event_record_types)

        # assign unique IDs
        for index, ert in enumerate(sorted(self._event_record_types, key=lambda evt: evt.name)):
            assert ert._id is None
            ert._id = Id(index)

        self._set_features(features)
        self._packet_context_field_type_extra_members = StructureFieldTypeMembers({})

        if packet_context_field_type_extra_members is not None:
            self._packet_context_field_type_extra_members = StructureFieldTypeMembers(packet_context_field_type_extra_members)

        self._set_pkt_ctx_ft()
        self._set_er_header_ft()

    def _set_features(self, features: Optional[DataStreamTypeFeatures]):
        if features is not None:
            self._features = features
            return None

        er_ts_ft = None
        pkt_beginning_ts_ft = None
        pkt_end_ts_ft = None

        if self._default_clock_type is not None:
            # Automatic timestamp field types because the data stream
            # type has a default clock type.
            er_ts_ft = DEFAULT_FIELD_TYPE
            pkt_beginning_ts_ft = DEFAULT_FIELD_TYPE
            pkt_end_ts_ft = DEFAULT_FIELD_TYPE

        self._features = DataStreamTypeFeatures(DataStreamTypePacketFeatures(beginning_timestamp_field_type=pkt_beginning_ts_ft,
                                                                             end_timestamp_field_type=pkt_end_ts_ft),
                                                DataStreamTypeEventRecordFeatures(timestamp_field_type=er_ts_ft))

    def _set_ft_mapped_clk_type_name(self, ft: Optional[UnsignedIntegerFieldType]):
        if ft is None:
            return

        if self._default_clock_type is not None:
            assert isinstance(ft, UnsignedIntegerFieldType)
            ft._mapped_clk_type_name = self._default_clock_type.name

    def _set_pkt_ctx_ft(self):
        members = None

        def add_member_if_exists(name: str, ft: _FieldType, set_mapped_clk_type_name: bool = False):
            nonlocal members

            if ft is not None:
                if set_mapped_clk_type_name:
                    self._set_ft_mapped_clk_type_name(typing.cast(UnsignedIntegerFieldType, ft))

                members[name] = StructureFieldTypeMember(ft)

        members = collections.OrderedDict([
            (
                'packet_size',
                StructureFieldTypeMember(self._features.packet_features.total_size_field_type)
            ),
            (
                'content_size',
                StructureFieldTypeMember(self._features.packet_features.content_size_field_type)
            )
        ])

        add_member_if_exists('timestamp_begin',
                             self._features.packet_features.beginning_timestamp_field_type, True)
        add_member_if_exists('timestamp_end', self._features.packet_features.end_timestamp_field_type,
                             True)
        add_member_if_exists('events_discarded',
                             self._features.packet_features.discarded_event_records_snapshot_counter_field_type)
        add_member_if_exists('packet_seq_num',
                             self._features.packet_features.sequence_number_field_type)

        if self._packet_context_field_type_extra_members is not None:
            for name, field_type in self._packet_context_field_type_extra_members.items():
                assert name not in members
                members[name] = field_type

        self._pkt_ctx_ft = StructureFieldType(8, members)

    def _set_er_header_ft(self):
        members = collections.OrderedDict()

        if self._features.event_record_features.type_id_field_type is not None:
            members['id'] = StructureFieldTypeMember(self._features.event_record_features.type_id_field_type)

        if self._features.event_record_features.timestamp_field_type is not None:
            ft = self._features.event_record_features.timestamp_field_type
            self._set_ft_mapped_clk_type_name(ft)
            members['timestamp'] = StructureFieldTypeMember(ft)

        self._er_header_ft = StructureFieldType(8, members)

    @property
    def id(self) -> Optional[Id]:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def default_clock_type(self) -> Optional[ClockType]:
        return self._default_clock_type

    @property
    def features(self) -> DataStreamTypeFeatures:
        return self._features

    @property
    def packet_context_field_type_extra_members(self) -> StructureFieldTypeMembers:
        return self._packet_context_field_type_extra_members

    @property
    def event_record_common_context_field_type(self) -> _OptStructFt:
        return self._event_record_common_context_field_type

    @property
    def event_record_types(self) -> FrozenSet[EventRecordType]:
        return self._event_record_types


_OptUuidFt = Optional[Union[str, StaticArrayFieldType]]


class TraceTypeFeatures:
    def __init__(self, magic_field_type: _OptDefaultableUIntFt = DEFAULT_FIELD_TYPE,
                 uuid_field_type: _OptUuidFt = None,
                 data_stream_type_id_field_type: _OptDefaultableUIntFt = DEFAULT_FIELD_TYPE):
        def get_field_type(user_ft: Optional[Union[str, _FieldType]],
                           create_default_ft: Callable[[], _FieldType]) -> _OptFt:
            if user_ft == DEFAULT_FIELD_TYPE:
                return create_default_ft()

            return typing.cast(_OptFt, user_ft)

        def create_default_magic_ft():
            return UnsignedIntegerFieldType(32)

        def create_default_uuid_ft():
            return StaticArrayFieldType(Count(16), UnsignedIntegerFieldType(8))

        def create_default_dst_id_ft():
            return UnsignedIntegerFieldType(64)

        self._magic_field_type = typing.cast(_OptUIntFt, get_field_type(magic_field_type, create_default_magic_ft))
        self._uuid_field_type = typing.cast(Optional[StaticArrayFieldType],
                                            get_field_type(uuid_field_type, create_default_uuid_ft))
        self._data_stream_type_id_field_type = typing.cast(_OptUIntFt,
                                                           get_field_type(data_stream_type_id_field_type,
                                                                          create_default_dst_id_ft))

    @property
    def magic_field_type(self) -> _OptUIntFt:
        return self._magic_field_type

    @property
    def uuid_field_type(self) -> Optional[StaticArrayFieldType]:
        return self._uuid_field_type

    @property
    def data_stream_type_id_field_type(self) -> _OptUIntFt:
        return self._data_stream_type_id_field_type


class _TraceType:
    def __init__(self, trace_byte_order: ByteOrder, data_stream_types: Set[DataStreamType],
                 uuid: _OptUuid, features: Optional[TraceTypeFeatures]):
        self._trace_byte_order = trace_byte_order
        self._data_stream_types = frozenset(data_stream_types)

        # assign unique IDs
        for index, dst in enumerate(sorted(self._data_stream_types, key=lambda st: st.name)):
            assert dst._id is None
            dst._id = Id(index)

        self._uuid = uuid
        self._set_features(features)
        self._set_pkt_header_ft()

    def _set_features(self, features: Optional[TraceTypeFeatures]):
        if features is not None:
            self._features = features
            return

        # automatic UUID field type because the trace type has a UUID
        uuid_ft = None if self._uuid is None else DEFAULT_FIELD_TYPE
        self._features = TraceTypeFeatures(uuid_field_type=uuid_ft)

    def _set_pkt_header_ft(self):
        members = collections.OrderedDict()

        def add_member_if_exists(name: str, ft: _OptFt):
            nonlocal members

            if ft is not None:
                members[name] = StructureFieldTypeMember(ft)

        add_member_if_exists('magic', self._features.magic_field_type)
        add_member_if_exists('uuid', self._features.uuid_field_type)
        add_member_if_exists('stream_id', self._features.data_stream_type_id_field_type)
        self._pkt_header_ft = StructureFieldType(8, members)

    @property
    def trace_byte_order(self) -> ByteOrder:
        return self._trace_byte_order

    @property
    def uuid(self) -> _OptUuid:
        return self._uuid

    @property
    def data_stream_types(self) -> FrozenSet[DataStreamType]:
        return self._data_stream_types

    def data_stream_type(self, name: str) -> Optional[DataStreamType]:
        for cand_dst in self._data_stream_types:
            if cand_dst.name == name:
                return cand_dst

        return None

    @property
    def features(self) -> TraceTypeFeatures:
        return self._features

    @property
    def clock_types(self) -> Set[ClockType]:
        clk_types = set()

        for dst in self._data_stream_types:
            if dst.default_clock_type is not None:
                clk_types.add(dst.default_clock_type)

        return clk_types


# Standard trace type class for a barectf 3 configuration.
#
# The trace byte order property of an instance is the same as the
# expected native byte order of the target system.
class TraceType(_TraceType):
    def __init__(self, trace_byte_order: ByteOrder, data_stream_types: Set[DataStreamType],
                 uuid: _OptUuid = None, features: Optional[TraceTypeFeatures] = None):
        super().__init__(trace_byte_order, data_stream_types, uuid, features)

    @property
    def native_byte_order(self) -> ByteOrder:
        return self._trace_byte_order


# This trace type class specifically exists to support barectf 2
# configurations. Such configurations define the actual trace byte order
# instead of the expected native byte order of the target system.
#
# The user provides the trace byte order property of an instance.
#
# IMPORTANT: When possible, prefer the `TraceType` class instead, as
# enforcing the trace byte order could result in a less efficient
# generated tracer.
class TraceTypeWithUnknownNativeByteOrder(_TraceType):
    def __init__(self, trace_byte_order: ByteOrder, data_stream_types: Set[DataStreamType],
                 uuid: _OptUuid = None, features: Optional[TraceTypeFeatures] = None):
        super().__init__(trace_byte_order, data_stream_types, uuid, features)


_EnvEntry = Union[str, int]
_EnvEntries = Mapping[str, _EnvEntry]


class TraceEnvironment(collections.abc.Mapping):
    def __init__(self, environment: _EnvEntries):
        self._env = {name: value for name, value in environment.items()}

    def __getitem__(self, key: str) -> _EnvEntry:
        return self._env[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._env)

    def __len__(self) -> int:
        return len(self._env)


class Trace:
    def __init__(self, type: _TraceType, environment: Optional[_EnvEntries] = None):
        self._type = type
        self._set_env(environment)

    def _set_env(self, environment: Optional[_EnvEntries]):
        init_env = collections.OrderedDict([
            ('domain', 'bare'),
            ('tracer_name', 'barectf'),
            ('tracer_major', barectf_version.__major_version__),
            ('tracer_minor', barectf_version.__minor_version__),
            ('tracer_patch', barectf_version.__patch_version__),
            ('tracer_pre', barectf_version.__pre_version__),
            ('barectf_gen_date', str(datetime.datetime.now().isoformat())),
        ])

        if environment is None:
            environment = {}

        init_env.update(environment)
        self._env = TraceEnvironment(typing.cast(_EnvEntries, init_env))

    @property
    def type(self) -> _TraceType:
        return self._type

    @property
    def environment(self) -> TraceEnvironment:
        return self._env


_ClkTypeCTypes = Mapping[ClockType, str]


class ClockTypeCTypes(collections.abc.Mapping):
    def __init__(self, c_types: _ClkTypeCTypes):
        self._c_types = {clk_type: c_type for clk_type, c_type in c_types.items()}

    def __getitem__(self, key: ClockType) -> str:
        return self._c_types[key]

    def __iter__(self) -> Iterator[ClockType]:
        return iter(self._c_types)

    def __len__(self) -> int:
        return len(self._c_types)


class ConfigurationCodeGenerationHeaderOptions:
    def __init__(self, identifier_prefix_definition: bool = False,
                 default_data_stream_type_name_definition: bool = False):
        self._identifier_prefix_definition = identifier_prefix_definition
        self._default_data_stream_type_name_definition = default_data_stream_type_name_definition

    @property
    def identifier_prefix_definition(self) -> bool:
        return self._identifier_prefix_definition

    @property
    def default_data_stream_type_name_definition(self) -> bool:
        return self._default_data_stream_type_name_definition


class ConfigurationCodeGenerationOptions:
    def __init__(self, identifier_prefix: str = 'barectf_', file_name_prefix: str = 'barectf',
                 default_data_stream_type: Optional[DataStreamType] = None,
                 header_options: Optional[ConfigurationCodeGenerationHeaderOptions] = None,
                 clock_type_c_types: Optional[_ClkTypeCTypes] = None):
        self._identifier_prefix = identifier_prefix
        self._file_name_prefix = file_name_prefix
        self._default_data_stream_type = default_data_stream_type

        self._header_options = ConfigurationCodeGenerationHeaderOptions()

        if header_options is not None:
            self._header_options = header_options

        self._clock_type_c_types = ClockTypeCTypes({})

        if clock_type_c_types is not None:
            self._clock_type_c_types = ClockTypeCTypes(clock_type_c_types)

    @property
    def identifier_prefix(self) -> str:
        return self._identifier_prefix

    @property
    def file_name_prefix(self) -> str:
        return self._file_name_prefix

    @property
    def default_data_stream_type(self) -> Optional[DataStreamType]:
        return self._default_data_stream_type

    @property
    def header_options(self) -> ConfigurationCodeGenerationHeaderOptions:
        return self._header_options

    @property
    def clock_type_c_types(self) -> ClockTypeCTypes:
        return self._clock_type_c_types


class ConfigurationOptions:
    def __init__(self,
                 code_generation_options: Optional[ConfigurationCodeGenerationOptions] = None):
        self._code_generation_options = ConfigurationCodeGenerationOptions()

        if code_generation_options is not None:
            self._code_generation_options = code_generation_options

    @property
    def code_generation_options(self) -> ConfigurationCodeGenerationOptions:
        return self._code_generation_options


class Configuration:
    def __init__(self, trace: Trace, options: Optional[ConfigurationOptions] = None):
        self._trace = trace
        self._options = ConfigurationOptions()

        if options is not None:
            self._options = options

        clk_type_c_types = self._options.code_generation_options.clock_type_c_types

        for dst in trace.type.data_stream_types:
            def_clk_type = dst.default_clock_type

            if def_clk_type is None:
                continue

            if def_clk_type not in clk_type_c_types:
                clk_type_c_types._c_types[def_clk_type] = 'uint32_t'

    @property
    def trace(self) -> Trace:
        return self._trace

    @property
    def options(self) -> ConfigurationOptions:
        return self._options
